#include <stdio.h>
/* based off of lin-bairstow.zeno, which I got here: 
 * http://home.att.net/~srschmitt/linbairstow.html
 */

#include <math.h>
#include <stdlib.h>
#include "lin-bairstow.h"


/*
%
% The Lin-Bairstow method improves on the deflation method to solve 
% for the complex roots of the following polynomial:
%
%     y = coeff(0) + coeff(1) X + coeff(2) X^2 +...+ coeff(n) X^n
%
% parameters:
%
%   coeff      must be an array with at least order+1 elements.
%   roots      output array of roots
%   order      order of polynomial
%
*/

int lin_bairstow(const double coeff[], cplx roots[], int order)
{
    double a[DIM], b[DIM], c[DIM], d[DIM];
    double alfa1, alfa2;
    double beta1, beta2;
    double delt1, delt2, delt3;
    int    i, j, k;
    int    count;
    int    n;
    int    steps;

    n  = order;
  
    /* is the coefficient of the highest term zero? */
    /* JZ:  Should this be coeff[order] instead?    */
    if (fabs( coeff[order] ) < SMALL)
        return FALSE;

    for (i = 0; i <= n; i++)
        a[n + 1 - i] = coeff[i];
    
    /* is highest coeff not close to 1? */
    if (fabs(a[1] - 1.0) > SMALL)
    {
        /* adjust coefficients because a(1) != 1 */
        for (i = 2; i <= n + 1; i++)
            a[i] /= a[1];

        a[1] = 1;
    }

    /* initialize root counter */
    count = 0;

    /*
    % start the main Lin-Bairstow iteration loop
    % initialize the counter and guesses for the
    % coefficients of quadratic factor:
    %
    % p(x) = x^2 + alfa1 * x + beta1
    */
    do
    {
        /* these are just guesses */
        alfa1 = 1;
        beta1 = 2;
        /* JZ:  Use a_n-1/a_n and a_n-2/a_n instead, as per Wikipedia? */
        /* Seems to actually take more steps with trivial examples at least. */
#   if 0
        alfa1 = a[n-0] / a[n + 1];
        beta1 = a[n-1] / a[n + 1];
#   endif
        steps = 0;
        do
        {
            b[0] = d[0] = 0;
            b[1] = d[1] = 1;
            j = 1;
            k = 0;
            for (i = 2; i <= n + 1; i++)
            {
                b[i] = a[i] - alfa1 * b[j] - beta1 * b[k];
                d[i] = b[i] - alfa1 * d[j] - beta1 * d[k];
                j++;
                k++;
            }
            j = n - 1;
            k = n - 2;

            delt1 = d[j] * d[j] - ( d[n] - b[n] ) * d[k];
            if (fabs(delt1) < SMALL)
                return FALSE;   /* cannot divide by zero */

            alfa2 = ( b[n  ] * d[j] - b[n+1] * d[k] ) / delt1;
            beta2 = ( b[n+1] * d[j] - ( d[n] - b[n] ) * b[n] ) / delt1;

            if (steps++ < 1000)
            {
                alfa1 = alfa1 + alfa2;
                beta1 = beta1 + beta2;
            } else
            {
                steps = 0;
                alfa1 = rand();
                beta1 = rand();
            }
        } while (fabs(alfa2) > TOLERANCE  || fabs(beta2) > TOLERANCE);

        delt1 = alfa1 * alfa1 - 4. * beta1;

        if (delt1 < 0)          /* imaginary roots */
        {
            delt2 = sqrt( fabs( delt1 ) ) / 2.;
            delt3 = -alfa1 / 2.;
            roots[count  ].re =  delt3;
            roots[count  ].im = +delt2;
            roots[count+1].re =  delt3;
            roots[count+1].im = -delt2;
        }
        else                    /* roots are real */
        {
            delt2 = sqrt( delt1 );
            roots[count  ].re = ( delt2 - alfa1 ) / 2;
            roots[count  ].im = 0;
            roots[count+1].re = ( delt2 + alfa1 ) / -2;
            roots[count+1].im = 0;
        }
        
        /* Update root counter and reduce polynomial order */
        count += 2;
        n     -= 2;

        /* for n >= 2 calculate coefficients of the new polynomial */
        if (n >= 2) 
            for (i = 1; i <= n+1; i++)
                a[i] = b[i];
    } while (n > 1);

    /* obtain last single real root */
    if (n == 1) 
    {
        roots[count].re = -b[2];
        roots[count].im = 0;
    }

    return TRUE;
}


#ifdef TEST
#include <stdio.h>

int main(void)
{
    cplx   roots[DIM];
    double poly [DIM] = { -12, 10, 4, 1, -4, 1 };
    double poly2[DIM] = { 1, -4, 1, 4, 10, -12 };
    double poly3[DIM] = { 6, 11, -33, -33, 11, 6 };
    int i;

    if (lin_bairstow(poly, roots, 5))
    {
        printf("Roots:\n");

        for (i = 0; i < 5; i++)
            printf("re=%f im=%f\n", roots[i].re, roots[i].im);
    } else
        printf("Didn't converge\n");

    if (lin_bairstow(poly2, roots, 5))
    {
        printf("Roots:\n");

        for (i = 0; i < 5; i++)
            printf("re=%f im=%f\n", roots[i].re, roots[i].im);
    } else
        printf("Didn't converge\n");

    
    if (lin_bairstow(poly3, roots, 5))
    {
        printf("Roots:\n");

        for (i = 0; i < 5; i++)
            printf("re=%f im=%f\n", roots[i].re, roots[i].im);
    } else
        printf("Didn't converge\n");

    return 0;
}
#endif
