/* ======================================================================== */
/*  Simple reencoder from LPC-10e output to Intellivoice, quick-and-dirty   */
/*                                                                          */
/*  As a proof of concept, this code takes the LPC-10e data output by       */
/*  the DoD LPC-10e encoder, and converts it (with minimal translation)     */
/*  to a series of uncompressed SP-0256 data blocks.                        */
/*                                                                          */
/*  LPC-10e outputs the following information for 180-sample frames.        */
/*                                                                          */
/*  1.  Pitch period in samples (quantized by tau table, unfortunately)     */
/*                                                                          */
/*  2.  Reflection coefficients for a 10th order lattice filter IIR.        */
/*                                                                          */
/*  3.  RMS energy of the block (ie. "amplitude").  Note that this isn't    */
/*      the excitation amplitude, but rather amplitude of the output.       */
/*      The two can be quite different if a given frame has a large         */
/*      resonance around the excitation frequency.                          */
/*                                                                          */
/*  That's not what the Intellivoice needs whatsoever!  What it needs is:   */
/*                                                                          */
/*  1.  Pitch period:  This is actually OK as is.  It's the only such       */
/*      parameter.  Do need to compute number of repeats though.            */
/*                                                                          */
/*  2.  "Frequency" and "Bandwidth" coefficients for its 2nd order IIR      */
/*      sections.  To get these, we need to go through multiple steps:      */
/*                                                                          */
/*      a.  Reverse the reflection coefficients into a direct-form IIR.     */
/*          The irc2pc function from the LPC-10e coder itself does this.    */
/*                                                                          */
/*      b.  Factor the direct-form IIR to find its roots (poles).           */
/*                                                                          */
/*      c.  Sort the poles:   (TO DO LATER)                                 */
/*           -- complex conjugate pairs by magnitude, largest last.         */
/*           -- real poles in descending order.                             */
/*                                                                          */
/*      d.  Go down the list converting pairs of poles into F and B         */
/*          terms to fit the poly 1 + 2*F * z^-1 + B * z^-2.                */
/*                                                                          */
/*      e.  Quantize F/B.                                                   */
/*                                                                          */
/*  3.  Amplitude of the excitation.  Since the IIR could have noticeable   */
/*      gain due to resonance, we need to run the IIR locally for one       */
/*      pitch period with an unit impulse, calculate up the RMS of the      */
/*      output, divide into the target RMS, and scale the amplitude         */
/*      accordingly.  WHEE!                                                 */
/*                                                                          */
/*  Once we have all this converted, we need to encode it for the SP0256.   */
/*  For now, we can use the braindead "Load All" opcode, and pack all the   */
/*  data into DECLEs.                                                       */
/*                                                                          */
/* ======================================================================== */


#include "config.h"

#include <stdint.h>
#include "irc2pc.h"
#include "lin-bairstow.h"
#include "lpc10/lpc10.h"

#define MAX_ORDER (12)  /* must be even */

/* ======================================================================== */
/*  Pitch map from LPC-10 encoder                                           */
/* ======================================================================== */
const int tau[61] = 
{ 
    20,  
    20,  21,  22,  23,  24,  25,  26,  27,  28,  29,  30,  31,
    32,  33,  34,  35,  36,  37,  38,  39,  40,  42,  44,  46,  48,
    50,  52,  54,  56,  58,  60,  62,  64,  66,  68,  70,  72,  74,
    76,  78,  80,  84,  88,  92,  96,  100, 104, 108, 112, 116,
    120, 124, 128, 132, 136, 140, 144, 148, 152, 156
};

/* ======================================================================== */
/*  F/B inverse quantization table                                          */
/*  Direct from the SP-0250 data sheet.                                     */
/* ======================================================================== */
const int16_t inv_quant_tbl[128] =
{
    0,      9,      17,     25,     33,     41,     49,     57,
    65,     73,     81,     89,     97,     105,    113,    121,
    129,    137,    145,    153,    161,    169,    177,    185,
    193,    201,    209,    217,    225,    233,    241,    249,
    257,    265,    273,    281,    289,    297,    301,    305,
    309,    313,    317,    321,    325,    329,    333,    337,
    341,    345,    349,    353,    357,    361,    365,    369,
    373,    377,    381,    385,    389,    393,    397,    401,
    405,    409,    413,    417,    421,    425,    427,    429,
    431,    433,    435,    437,    439,    441,    443,    445,
    447,    449,    451,    453,    455,    457,    459,    461,
    463,    465,    467,    469,    471,    473,    475,    477,
    479,    481,    482,    483,    484,    485,    486,    487,
    488,    489,    490,    491,    492,    493,    494,    495,
    496,    497,    498,    499,    500,    501,    502,    503,
    504,    505,    506,    507,    508,    509,    510,    511
};

/* ======================================================================== */
/*  F/B quantization table:  Computed from inverse quant table, rounding    */
/*  down in all cases.                                                      */
/* ======================================================================== */
int16_t quant_tbl[512];

static void init_quant_table(void)
{
    static int init = 0;
    int i, j;

    if (init) return;
    init = 1;

    for (i = j = 0; i < 127; i++)
        while (j < inv_quant_tbl[i+1])
            quant_tbl[j++] = i;

    while (j < 512)
        quant_tbl[j++] = i;
}


/* ======================================================================== */
/*  Quantize double                                                         */
/* ======================================================================== */
LOCAL uint32_t quantize_double(double f)
{
    int fxpt, fxqt;

    init_quant_table();

    fxpt = 512 * fabs(f);

    fxqt = (fxpt > 511) ? 127 : quant_tbl[fxpt];

    fxqt -= 2;
    if (fxqt < 0) fxqt = 0;

    return f > 0 ? 0xFF & -fxqt : fxqt;
}

/* ======================================================================== */
/*  NEXT_RNG_SAMP                                                           */
/* ======================================================================== */
LOCAL double next_rng_samp(uint32_t *seed)
{
    *seed = (*seed >> 1) ^ (*seed & 1 ? 0x4001 : 0);
    return (*seed) & 1 ? 1. : -1.;
}


/* ======================================================================== */
/*  Determine amplitude                                                     */
/* ======================================================================== */
LOCAL int determine_amplitude
(
    const double f[MAX_ORDER / 2], 
    const double b[MAX_ORDER / 2],
    int   period,
    int   voiced,
    double target_rms,
    double gain_scale
)
{
    int i, j;
    double delay[MAX_ORDER / 2][2];
    double target_sum_sq, actual_sum_sq, gain, samp;
    uint32_t seed = 0x2A3A4A5A;

    for (i = 0; i < 6; i++)
        delay[i][0] = delay[i][1] = 0;

    /* -------------------------------------------------------------------- */
    /*  Run the synthesis filter for "period" samples to see what we get.   */
    /* -------------------------------------------------------------------- */
    actual_sum_sq = 0.;

    for (i = 0; i < period; i++)
    {
        samp = voiced ? !i : next_rng_samp(&seed);

        for (j = 0; j < MAX_ORDER / 2; j++)
        {
            samp +=       b[j] * delay[j][1];
            samp += 2.0 * f[j] * delay[j][0];

            delay[j][1] = delay[j][0];
            delay[j][0] = samp;
        }

        actual_sum_sq += samp * samp;
    }

    /* -------------------------------------------------------------------- */
    /*  Now compute gain by comparing actual sum of the squares to the      */
    /*  target RMS.  This comes straight from the LPC-10e FORTRAN code.     */
    /* -------------------------------------------------------------------- */
    target_sum_sq = target_rms * target_rms * (double)period;
    gain = sqrt(target_sum_sq / actual_sum_sq);

    /* -------------------------------------------------------------------- */
    /*  Convert the amplitude into a fixed point number, and return it.     */
    /* -------------------------------------------------------------------- */
    return (int)(gain * gain_scale);
}

/* ======================================================================== */
/*  ENCODE_AMPLITUDE                                                        */
/* ======================================================================== */
LOCAL uint32_t encode_amplitude(int ampl)
{
    int expn = 0;

    if (ampl < 0) ampl = -ampl;

    while (ampl > 0x1F)
    {
        ampl >>= 1;
        expn++;
    }

    expn <<= 5;

    if (expn > 0xE0) expn = 0xE0;

    expn -= 0x20;
    if (expn < 0) expn = 0;

    return expn | (ampl & 0x1F);
}

LOCAL double mag(cplx c)
{
    return sqrt(c.re*c.re + c.im*c.im);
}

/* ======================================================================== */
/*  RC to FB                                                                */
/* ======================================================================== */
LOCAL void rc_to_fb
(
    const double rc[MAX_ORDER],
    double f[MAX_ORDER / 2],
    double b[MAX_ORDER / 2]
)
{
    double pc[MAX_ORDER + 2];
    cplx   roots[14];
    int    order = 10;     /* later, will bump this to order 12 */
    int    lb_order;
    int    i, j;

    assert(order <= MAX_ORDER);

    /* -------------------------------------------------------------------- */
    /*  Assume that the filter is empty.                                    */
    /* -------------------------------------------------------------------- */
    for (i = 0; i < MAX_ORDER / 2; i++)
        f[i] = b[i] = 0;

    /* -------------------------------------------------------------------- */
    /*  IRC2PC will return pc[1] .. pc[order] as coefficients of the        */
    /*  polynomial:                                                         */
    /*                                                                      */
    /*      1 + pc[1] * z^-1 + pc[2] * z^-2 + ... + pc[order] * z^-order    */
    /*                                                                      */
    /*  It is that polynomial that we want to factor into 2nd-order blocks  */
    /*  of the form:                                                        */
    /*                                                                      */
    /*      1 + 2*F[stage] * z^-1 + B[stage] * z^-2                         */
    /* -------------------------------------------------------------------- */
    pc[0] = -1.0;
    irc2pc(rc, &pc[1], order);

    /* -------------------------------------------------------------------- */
    /*  Find the highest non-zero term, and take that as the actual order   */
    /*  of the polynomial.  We need this because our Lin-Bairstow impl.     */
    /*  expects us to.                                                      */
    /* -------------------------------------------------------------------- */
    for (i = 1, lb_order = 0; i <= order; i++)
        if (fabs(pc[i]) >= SMALL)
            lb_order = i;

    /* -------------------------------------------------------------------- */
    /*  Convert "1 + k0 * z^-1 + k1 * z^-2 + .. k9 * z^-10 into a poly w/   */
    /*  positive powers of Z:  z^10 + k0 * z^9 + k1 * z^8 + ... + k0        */
    /* -------------------------------------------------------------------- */
    for (i = 0, j = lb_order; i <= j; i++, j--)
    {
        double temp;
        temp  = pc[i];
        pc[i] = pc[j];
        pc[j] = temp;
    }

    /* -------------------------------------------------------------------- */
    /*  If there was at least 1 non-zero term in the coefficient set, then  */
    /*  go ahead and factor it.                                             */
    /* -------------------------------------------------------------------- */
    if (lb_order > 0)
        if (lin_bairstow(pc, roots, lb_order) == FALSE)
        {
            fprintf(stderr, "lb_order=%d\n", lb_order);
            for (i = 0; i < order; i += 2)
                fprintf(stderr, "%+-24.20f, %+-24.20f\n", pc[i], pc[i+1]);
            fprintf(stderr, "%+-24.20f\n", pc[order]);
            assert(0 && "lin_bairstow failed");
        }

    /* -------------------------------------------------------------------- */
    /*  For all the pairs of roots, reconstitute F and B.                   */
    /* -------------------------------------------------------------------- */
    for (i = 0; i < lb_order / 2; i++)
    {
        /* For now just do the same for all roots;  */
        f[i] = + 0.5 * (roots[2*i].re + roots[2*i+1].re);
        b[i] = - roots[2*i].re * roots[2*i + 1].re 
               + roots[2*i].im * roots[2*i + 1].im;
    }

    /* -------------------------------------------------------------------- */
    /*  Handle the oddball real root, if any.                               */
    /* -------------------------------------------------------------------- */
    if (lb_order & 1)
    {
        f[i] = 0.5 * roots[2*i].re;
        b[i] = 0.;
    }
}

/* ======================================================================== */
/*  READ_16BIT_SAMPLES_AS_FLOAT                                             */
/* ======================================================================== */
LOCAL int read_16bit_samples_as_float
(
    FILE  *RESTRICT f, 
    float *RESTRICT out, 
    int             count
)
{
    int16_t buf[256];
    int i, to_read, got, total = 0;

    while (count > 0)
    {
        to_read = count > 256 ? 256 : count;
        count  -= to_read;
        got     = fread(buf, sizeof(int16_t), to_read, f);

        if (got <= 0)
            break;

        for (i = 0; i < got; i++)
            out[total++] = LE_TO_HOST_16(buf[i]) / 32768.0f;

        if (got != to_read)
            break;
    }

    for (i = total; i < count; i++)
        out[i] = 0.f;

    return total;
}

/* ======================================================================== */
/*  USAGE                                                                   */
/* ======================================================================== */
LOCAL void usage(void)
{
        fprintf (stderr, 

"Usage: ivoice_enc foo.raw > foo.asm\n"
"\n"
"Input file 'foo.raw' should be a headerless 10kHz signed 16-bit audio file.\n"
"You can create such a file from other formats with the following sox\n"
"command line:\n"
"\n"
"    sox input.wav -t raw -s -2 -r 10000 --endian little output.raw rate\n"
"\n"
"where \"input.wav\" is the input file and \"output.raw\" is the output file."
"\n"
"\n"
        );
}


/* ======================================================================== */
/*  MAIN                                                                    */
/* ======================================================================== */
int main(int argc, char *argv[])
{
    double rc[12], f[6], b[6];
    int voiced;
    int period;
    int ampl, ampl_enc;
    double rms;
    char buf[1024];
    int i;
    uint32_t fi[6], bi[6];
    uint32_t out[14];
    FILE *f_in;

    float speech[LPC10_SAMPLES_PER_FRAME];
    lpc10_encoder_state  *st;
    lpc10_analyzed_frame fr;

    int frame_size  = 180;
    int frame_extra = 0;
    int repeat_cnt  = 0;
    int frame_no    = 0;
    int started     = 0;
    int silent      = 0;
    int tot_silence = 0;
    int odd = 0;

    if (argc != 2 || argv[1][0] == '-')
    {
        usage();
        exit(1);
    }
    if ((f_in = fopen(argv[1], "rb")) == NULL)
    {
        fprintf(stderr, "Could not open %s for reading\n", argv[1]);
        exit(1);
    }
    
    st = create_lpc10_encoder_state();

    while (read_16bit_samples_as_float(f_in, speech, 180) == 180)
    {
        fputc('.', stderr); 
        if ((++frame_no & 63) == 0) fputc('\n', stderr);

        lpc10_analyze_frame(speech, &fr, st);

        voiced = fr.voiced;
        period = tau[fr.pitch];
        rms    = fr.rms;

        for (i = 0; i < 5; i++)
        {
            rc[2*i + 0] = fr.rcoef[i][0];
            rc[2*i + 1] = fr.rcoef[i][1];
        }
        rc[11] = rc[12] = 0;

        if (voiced == 0)
        {
            period = 64;
            rms   *= 2;
        }

        rc_to_fb(rc, f, b);

        ampl     = determine_amplitude(f, b, period, voiced, rms, 1);
        ampl_enc = encode_amplitude(ampl);

        if (ampl < 32)
        {
            period = 64;
            silent = 1;
        } else
            silent = 0;


        for (i = 0; i < 6; i++)
        {
            fi[5-i] = quantize_double(f[i]);
            bi[5-i] = quantize_double(b[i]);
        }

        repeat_cnt  = (frame_size + frame_extra) / period;
        frame_extra = (frame_size + frame_extra) - (repeat_cnt * period);

        if (!silent)
        {
            if (!started)
            {
                tot_silence = 0;
                started     = 1;
            }

            if (tot_silence > 0)
            {
                while (tot_silence > 63)
                {
                    printf("    IV.PAUSE    $3F\n");
                    tot_silence -= 63;
                }
                printf("    IV.PAUSE    $%02X\n", tot_silence);
                tot_silence = 0;
            }

            printf("    IV.LOAD\n");
            printf("        IV.RPT  $%02X\n",   repeat_cnt);
            printf("        IV.AMP  $%02X\n",   ampl_enc);
            printf("        IV.PER  $%02X\n",   voiced ? period : 0);
            for (i = 0; i < 6; i++)
            {
                printf("        IV.F    %d, $%.2X\n", i, fi[5-i]);
                printf("        IV.B    %d, $%.2X\n", i, bi[5-i]);
            }
            printf("    IV.LOAD.done\n");  /* to do: get rid of this */
        } else
        {
            tot_silence += repeat_cnt;
        }
    }
    printf("    IV.PAUSE 1\n");
    printf("    IV.RTS\n");  /* to do: get rid of this */
    return 0;
}
